@@ -1,13 +1,14 @@
 from pathlib import Path
-from typing import List, Optional
+from tempfile import TemporaryDirectory
+from typing import Optional
 
-from rich.progress import Progress
 from typer import Argument, Option
 
 from belay import Device
-from belay.cli.common import help_password, help_port, load_toml
+from belay.cli.common import help_password, help_port
 from belay.cli.run import run as run_cmd
 from belay.cli.sync import sync
+from belay.project import find_project_folder, load_groups, load_pyproject
 
 
 def install(
@@ -26,28 +27,44 @@ def install(
         raise ValueError("Run script MUST be a python file.")
     if main and main.suffix != ".py":
         raise ValueError("Main script MUST be a python file.")
-    toml = load_toml()
-    pkg_name = toml.get("name")
 
-    sync(
-        port=port,
-        folder=Path(".belay/dependencies/main"),
-        dst="/lib",
-        password=password,
-        keep=None,
-        ignore=None,
-        mpy_cross_binary=mpy_cross_binary,
-    )
-    if pkg_name:
+    toml = load_pyproject()
+    project_folder = find_project_folder()
+    project_package = toml.get("name")
+    groups = load_groups()
+
+    with TemporaryDirectory() as tmp_dir:
+        # Aggregate dependencies to an intermediate temporary directory.
+        tmp_dir = Path(tmp_dir)
+
+        for group in groups:
+            if group.config.optional:
+                # TODO: Should only iterate over non-optional & manually specified groups.
+                raise NotImplementedError
+            group.copy_to(tmp_dir)
+
         sync(
             port=port,
-            folder=Path(pkg_name),
-            dst=f"/{pkg_name}",
+            folder=tmp_dir,
+            dst="/lib",
             password=password,
             keep=None,
             ignore=None,
             mpy_cross_binary=mpy_cross_binary,
         )
+
+    if project_package:
+        project_package = Path(project_package)
+        sync(
+            port=port,
+            folder=project_folder / project_package,
+            dst=f"/{project_package.name}",
+            password=password,
+            keep=None,
+            ignore=None,
+            mpy_cross_binary=mpy_cross_binary,
+        )
+
     if main:
         with Device(port, password=password) as device:
             device.sync(main, keep=True, mpy_cross_binary=mpy_cross_binary)
