@@ -1,71 +1,37 @@
 import ast
 import concurrent.futures
-import importlib.resources as pkg_resources
 import linecache
+import math
 import re
-import secrets
-import string
 import subprocess  # nosec
 import sys
 import tempfile
-from abc import ABC, abstractmethod
 from dataclasses import dataclass
-from functools import lru_cache, wraps
-from inspect import isgeneratorfunction, signature
+from inspect import signature
 from pathlib import Path
-from typing import Callable, Dict, Generator, List, Optional, Set, TextIO, Tuple, Union
+from threading import Lock
+from typing import Callable, Optional, TextIO, Tuple, Union
 
+from autoregistry import Registry
 from pathspec import PathSpec
 from serial import SerialException
 
-from . import snippets
 from ._minify import minify as minify_code
-from .exceptions import (
-    ConnectionLost,
-    FeatureUnavailableError,
-    MaxHistoryLengthError,
-    SpecialFunctionNameError,
+from .exceptions import ConnectionLost, MaxHistoryLengthError
+from .executers import (
+    Executer,
+    SetupExecuter,
+    TaskExecuter,
+    TeardownExecuter,
+    ThreadExecuter,
 )
-from .inspect import getsource
+from .hash import fnv1a
+from .helpers import read_snippet, wraps_partial
+from .inspect import isexpression
 from .pyboard import Pyboard, PyboardError, PyboardException
+from .typing import BelayReturn
 from .webrepl import WebreplToSerial
 
-# Typing
-PythonLiteral = Union[None, bool, bytes, int, float, str, List, Dict, Set]
-BelayGenerator = Generator[PythonLiteral, None, None]
-BelayReturn = Union[BelayGenerator, PythonLiteral]
-BelayCallable = Callable[..., BelayReturn]
-
-
-_python_identifier_chars = (
-    string.ascii_uppercase + string.ascii_lowercase + string.digits
-)
-
-
-@lru_cache
-def _read_snippet(name):
-    return pkg_resources.read_text(snippets, f"{name}.py")
-
-
-def _local_hash_file(fn: Union[str, Path]) -> int:
-    """Compute the FNV-1a 32-bit hash of a file."""
-    fn = Path(fn)
-    h = 0x811C9DC5
-    size = 1 << 32
-    with fn.open("rb") as f:
-        while True:
-            data = f.read(65536)
-            if not data:
-                break
-            for byte in data:
-                h = h ^ byte
-                h = (h * 0x01000193) % size
-    return h
-
-
-def _random_python_identifier(n=16):
-    return "_" + "".join(secrets.choice(_python_identifier_chars) for _ in range(n))
-
 
 class NotBelayResponse(Exception):
     """Parsed response wasn't for Belay."""
@@ -87,224 +53,6 @@ def _parse_belay_response(line):
         raise ValueError(f'Received unknown code: "{code}"')
 
 
-class _Executer(ABC):
-    def __init__(self, device):
-        # To avoid Executer.__setattr__ raising an error
-        object.__setattr__(self, "_belay_device", device)
-
-    def __setattr__(self, name: str, value: BelayCallable):
-        if (
-            name.startswith("_belay")
-            or name.startswith("__belay")
-            or (name.startswith("__") and name.endswith("__"))
-        ):
-            raise SpecialFunctionNameError(
-                f'Not allowed to register function named "{name}".'
-            )
-        super().__setattr__(name, value)
-
-    def __getattr__(self, name: str) -> BelayCallable:
-        # Just here for linting purposes.
-        raise AttributeError
-
-    @abstractmethod
-    def __call__(self):
-        raise NotImplementedError
-
-
-class _TaskExecuter(_Executer):
-    def __call__(
-        self,
-        f: Optional[BelayCallable] = None,
-        /,
-        minify: bool = True,
-        register: bool = True,
-        record: bool = False,
-    ) -> BelayCallable:
-        """Decorator that send code to device that executes when decorated function is called on-host.
-
-        Parameters
-        ----------
-        f: Callable
-            Function to decorate. Can only accept and return python literals.
-        minify: bool
-            Minify ``cmd`` code prior to sending.
-            Defaults to ``True``.
-        register: bool
-            Assign an attribute to ``self`` with same name as ``f``.
-            Defaults to ``True``.
-        record: bool
-            Record task execution calls for state-reconstruction if device is accidentally reset.
-            Only recommended for tasks that are called a low amount of times to setup device state.
-            Defaults to ``False``.
-
-        Returns
-        -------
-        Callable
-            Remote-executor function.
-        """
-        if f is None:
-            return self  # type: ignore
-
-        name = f.__name__
-        src_code, src_lineno, src_file = getsource(f)
-        src_lineno -= 1  # Because of the injected ``@__belay`` decorator below
-
-        # Add the __belay decorator for handling result serialization.
-        src_code = f"@__belay({repr(name)})\n" + src_code
-
-        # Send the source code over to the device.
-        self._belay_device(src_code, minify=minify)
-
-        @wraps(f)
-        def func_executer(*args, **kwargs):
-            cmd = f"_belay_{name}(*{repr(args)}, **{repr(kwargs)})"
-
-            return self._belay_device._traceback_execute(
-                src_file, src_lineno, name, cmd, record=record
-            )
-
-        @wraps(f)
-        def multi_func_executer(*args, **kwargs):
-            res = func_executer(*args, **kwargs)
-            if hasattr(f, "_belay_level"):
-                # Call next device's wrapper.
-                if f._belay_level == 1:
-                    res = [f(*args, **kwargs), res]
-                else:
-                    res = [*f(*args, **kwargs), res]
-
-            return res
-
-        multi_func_executer._belay_level = 1
-        if hasattr(f, "_belay_level"):
-            multi_func_executer._belay_level += f._belay_level
-
-        @wraps(f)
-        def gen_executer(*args, **kwargs):
-            if record:
-                raise NotImplementedError(
-                    "Recording of generator tasks is currently not supported."
-                )
-            # Step 1: Create the on-device generator
-            gen_identifier = _random_python_identifier()
-            cmd = f"{gen_identifier} = _belay_{name}(*{repr(args)}, **{repr(kwargs)})"
-            self._belay_device._traceback_execute(
-                src_file, src_lineno, name, cmd, record=False
-            )
-            # Step 2: Create the host generator that invokes ``next()`` on-device.
-
-            def gen_inner():
-                send_val = None
-                try:
-                    while True:
-                        cmd = f"__belay_gen_next({gen_identifier}, {repr(send_val)})"
-                        send_val = yield self._belay_device._traceback_execute(
-                            src_file, src_lineno, name, cmd, record=False
-                        )
-                except StopIteration:
-                    pass
-                # Delete the exhausted generator on-device.
-                self._belay_device(f"del {gen_identifier}")
-
-            return gen_inner()
-
-        @wraps(f)
-        def multi_gen_executer(*args, **kwargs):
-            raise NotImplementedError
-
-        multi_gen_executer._belay_level = 1
-        if hasattr(f, "_belay_level"):
-            multi_gen_executer._belay_level += f._belay_level
-
-        if isgeneratorfunction(f):
-            executer = gen_executer
-
-            # TODO: define multi_gen_executer
-            if multi_gen_executer._belay_level > 1:
-                raise NotImplementedError(
-                    "Multi-device generator task decorating not yet implemented."
-                )
-        else:
-            executer = multi_func_executer
-
-        if register:
-            setattr(self, name, executer)
-
-        return executer
-
-
-class _ThreadExecuter(_Executer):
-    def __call__(
-        self,
-        f: Optional[Callable[..., None]] = None,
-        /,
-        minify: bool = True,
-        register: bool = True,
-        record: bool = True,
-    ) -> Callable[..., None]:
-        """Decorator that send code to device that spawns a thread when executed.
-
-        Parameters
-        ----------
-        f: Callable
-            Function to decorate. Can only accept python literals as arguments.
-        minify: bool
-            Minify ``cmd`` code prior to sending.
-            Defaults to ``True``.
-        register: bool
-            Assign an attribute to ``self`` with same name as ``f``.
-            Defaults to ``True``.
-        record: bool
-            Record thread execution calls for state-reconstruction if device is accidentally reset.
-            Defaults to ``True``.
-
-        Returns
-        -------
-        Callable
-            Remote-executor function.
-        """
-        if f is None:
-            return self  # type: ignore
-
-        if self._belay_device.implementation.name == "circuitpython":
-            raise FeatureUnavailableError("CircuitPython does not support threading.")
-
-        name = f.__name__
-        src_code, src_lineno, src_file = getsource(f)
-
-        # Send the source code over to the device.
-        self._belay_device(src_code, minify=minify)
-
-        @wraps(f)
-        def executer(*args, **kwargs):
-            cmd = f"import _thread; _thread.start_new_thread({name}, {repr(args)}, {repr(kwargs)})"
-            self._belay_device._traceback_execute(
-                src_file, src_lineno, name, cmd, record=record
-            )
-
-        @wraps(f)
-        def multi_executer(*args, **kwargs):
-            res = executer(*args, **kwargs)
-            if hasattr(f, "_belay_level"):
-                # Call next device's wrapper.
-                if f._belay_level == 1:
-                    res = [f(*args, **kwargs), res]
-                else:
-                    res = [*f(*args, **kwargs), res]
-
-            return res
-
-        multi_executer._belay_level = 1
-        if hasattr(f, "_belay_level"):
-            multi_executer._belay_level += f._belay_level
-
-        if register:
-            setattr(self, name, executer)
-
-        return multi_executer
-
-
 def _discover_files_dirs(
     remote_dir: str,
     local_file_or_folder: Path,
@@ -356,7 +104,7 @@ def _preprocess_keep(
         keep = []
     else:
         raise ValueError
-    keep = [str(dst / Path(x)) for x in keep]
+    keep = [(dst / Path(x)).as_posix() for x in keep]
     return keep
 
 
@@ -404,12 +152,12 @@ def _preprocess_src_file(
 
 def _preprocess_src_file_hash(*args, **kwargs):
     src_file = _preprocess_src_file(*args, **kwargs)
-    src_hash = _local_hash_file(src_file)
+    src_hash = fnv1a(src_file)
     return src_file, src_hash
 
 
 def _generate_dst_dirs(dst, src, src_dirs) -> list:
-    dst_dirs = [str(dst / x.relative_to(src)) for x in src_dirs]
+    dst_dirs = [(dst / x.relative_to(src)).as_posix() for x in src_dirs]
     # Add all directories leading up to ``dst``.
     dst_prefix_tokens = dst.split("/")
     for i in range(2, len(dst_prefix_tokens) + (dst[-1] != "/")):
@@ -418,6 +166,16 @@ def _generate_dst_dirs(dst, src, src_dirs) -> list:
     return dst_dirs
 
 
+def _sort_executers(executers):
+    def get_key(x):
+        try:
+            return x.__wrapped__.__belay__.id
+        except AttributeError:
+            return math.inf
+
+    return sorted(executers, key=get_key)
+
+
 @dataclass
 class Implementation:
     """Implementation dataclass detailing the device.
@@ -442,7 +200,25 @@ class Implementation:
     emitters: Tuple[str]
 
 
-class Device:
+_method_metadata_counter_lock = Lock()
+_method_metadata_counter = 0
+
+
+@dataclass
+class MethodMetadata:
+    executer: Callable
+    kwargs: dict
+    autoinit: bool = False  # Only applies to ``SetupExecuter``.
+    id: int = -1  # monotonically increasing global identifier.
+
+    def __post_init__(self):
+        global _method_metadata_counter
+        with _method_metadata_counter_lock:
+            self.id = _method_metadata_counter
+            _method_metadata_counter += 1
+
+
+class Device(Registry):
     """Belay interface into a micropython device.
 
     Attributes
@@ -476,18 +252,20 @@ class Device:
 
         self._connect_to_board(**self._board_kwargs)
 
-        self.task = _TaskExecuter(self)
-        self.thread = _ThreadExecuter(self)
+        for executer_name, executer_cls in Executer.items():
+            executer = executer_cls(self)
+            setattr(
+                self, executer_name, executer
+            )  # Public interface; might get stomped
+            setattr(
+                self, "_belay_" + executer_name, executer
+            )  # Private interface; will always be there
 
         self._exec_snippet("startup")
 
         self.implementation = Implementation(
             *self(
-                'print("_BELAYR("'
-                '+ repr(sys.implementation.name) + ","'
-                '+ repr(sys.implementation.version) + ","'
-                '+ repr(sys.platform) + ","'
-                '+")")'
+                "(sys.implementation.name, sys.implementation.version, sys.platform)"
             ),
             emitters=self._emitter_check(),
         )
@@ -500,6 +278,29 @@ class Device:
         elif startup:
             self(startup)
 
+        # if subclassing Device, register methods decorated with
+        # executer markers (e.g. ``Device.task``).
+        autoinit_executers = []
+        for name, method in vars(type(self)).items():
+            metadata = getattr(method, "__belay__", None)
+            if not metadata:
+                continue
+            decorator = getattr(self, metadata.executer.__registry__.name)
+            executer = decorator(method, **metadata.kwargs)
+
+            if metadata.autoinit:
+                autoinit_executers.append(executer)
+
+            setattr(
+                self,
+                name,
+                executer,
+            )
+
+        autoinit_executers = _sort_executers(autoinit_executers)
+        for executer in autoinit_executers:
+            executer()
+
     def _emitter_check(self):
         # Detect which emitters are available
         emitters = []
@@ -543,12 +344,13 @@ class Device:
         names : str
             Snippet(s) to load and execute.
         """
-        snippets = [_read_snippet(name) for name in names]
+        snippets = [read_snippet(name) for name in names]
         return self("\n".join(snippets))
 
     def __call__(
         self,
         cmd: str,
+        *,
         minify: bool = True,
         stream_out: TextIO = sys.stdout,
         record=True,
@@ -558,7 +360,7 @@ class Device:
         Parameters
         ----------
         cmd: str
-            Python code to execute.
+            Python code to execute. May be a statement or expression.
         minify: bool
             Minify ``cmd`` code prior to sending.
             Reduces the number of characters that need to be transmitted.
@@ -569,11 +371,15 @@ class Device:
 
         Returns
         -------
-            Return value from executing code on-device.
+            Correctly interpreted return value from executing code on-device.
         """
         if minify:
             cmd = minify_code(cmd)
 
+        if isexpression(cmd):
+            # Belay Tasks are inherently expressions as well.
+            cmd = f"print('_BELAYR' + repr({cmd}))"
+
         if (
             record
             and self.attempts
@@ -697,7 +503,7 @@ class Device:
                 dst_file.with_suffix(".mpy") if dst_file.suffix == ".py" else dst_file
                 for dst_file in dst_files
             ]
-        dst_files = [str(dst_file) for dst_file in dst_files]
+        dst_files = [dst_file.as_posix() for dst_file in dst_files]
         dst_dirs = _generate_dst_dirs(dst, folder, src_dirs)
 
         if keep_all:
@@ -750,10 +556,6 @@ class Device:
                 if progress_update:
                     progress_update(advance=1)
 
-        # Remove all the files and directories that did not exist in local filesystem.
-        if progress_update:
-            progress_update(description="Cleaning up...")
-
     def __enter__(self):
         return self
 
@@ -762,6 +564,10 @@ class Device:
 
     def close(self) -> None:
         """Close the connection to device."""
+        # Invoke all teardown executers prior to closing out connection.
+        for executer in _sort_executers(self._belay_teardown._belay_executers):
+            executer()
+
         return self._board.close()
 
     def reconnect(self, attempts: Optional[int] = None) -> None:
@@ -791,6 +597,140 @@ class Device:
         for cmd in self._cmd_history:
             self(cmd, record=False)
 
+    @staticmethod
+    def setup(f=None, *, autoinit=False, **kwargs) -> Callable:
+        """setup(f, *, minify=True, register=True, record=True)
+
+        Decorator that executes function's body in a global-context on-device when called.
+
+        Function arguments are also set in the global context.
+
+        Can either be used as a staticmethod ``@Device.setup`` for marking methods in a subclass of ``Device``, or as a standard method ``@device.setup`` for marking functions to a specific ``Device`` instance.
+
+        Parameters
+        ----------
+        f: Callable
+            Function to decorate. Can only accept and return python literals.
+        minify: bool
+            Minify ``cmd`` code prior to sending.
+            Defaults to ``True``.
+        register: bool
+            Assign an attribute to ``self.setup`` with same name as ``f``.
+            Defaults to ``True``.
+        record: bool
+            Each invocation of the executer is recorded for playback upon reconnect.
+            Defaults to ``True``.
+        autoinit: bool
+            Automatically invokes decorated functions at the end of object ``__init__``.
+            Methods will be executed in order-registered.
+            Defaults to ``False``.
+        """  # noqa: D400
+        if f is None:
+            return wraps_partial(Device.setup, autoinit=autoinit, **kwargs)  # type: ignore[reportGeneralTypeIssues]
+        if signature(f).parameters:
+            raise ValueError(
+                f"Method {f} decorated with "
+                '"@Device.setup(autoinit=True)" '
+                "must have no arguments."
+            )
+
+        f.__belay__ = MethodMetadata(
+            executer=SetupExecuter, autoinit=autoinit, kwargs=kwargs
+        )
+        return f
+
+    @staticmethod
+    def teardown(f=None, **kwargs) -> Callable:
+        """teardown(f, *, minify=True, register=True, record=True)
+
+        Decorator that executes function's body in a global-context on-device
+        when ``device.close()`` is called.
+
+        Function arguments are also set in the global context.
+
+        Can either be used as a staticmethod ``@Device.teardown`` for marking methods in a subclass of ``Device``, or as a standard method ``@device.teardown`` for marking functions to a specific ``Device`` instance.
+
+        Parameters
+        ----------
+        f: Callable
+            Function to decorate. Can only accept and return python literals.
+        minify: bool
+            Minify ``cmd`` code prior to sending.
+            Defaults to ``True``.
+        register: bool
+            Assign an attribute to ``self.teardown`` with same name as ``f``.
+            Defaults to ``True``.
+        record: bool
+            Each invocation of the executer is recorded for playback upon reconnect.
+            Defaults to ``True``.
+        """  # noqa: D400
+        if f is None:
+            return wraps_partial(Device.setup, **kwargs)  # type: ignore[reportGeneralTypeIssues]
+
+        if signature(f).parameters:
+            raise ValueError(
+                f'Method {f} decorated with "@Device.teardown" must have no arguments.'
+            )
+
+        f.__belay__ = MethodMetadata(executer=TeardownExecuter, kwargs=kwargs)
+        return f
+
+    @staticmethod
+    def task(f=None, **kwargs) -> Callable:
+        """task(f, *, minify=True, register=True, record=False)
+
+        Decorator that send code to device that executes when decorated function is called on-host.
+
+        Can either be used as a staticmethod ``@Device.task`` for marking methods in a subclass of ``Device``, or as a standard method ``@device.task`` for marking functions to a specific ``Device`` instance.
+
+        Parameters
+        ----------
+        f: Callable
+            Function to decorate. Can only accept and return python literals.
+        minify: bool
+            Minify ``cmd`` code prior to sending.
+            Defaults to ``True``.
+        register: bool
+            Assign an attribute to ``self.task`` with same name as ``f``.
+            Defaults to ``True``.
+        record: bool
+            Each invocation of the executer is recorded for playback upon reconnect.
+            Only recommended to be set to ``True`` for a setup-like function.
+            Defaults to ``False``.
+        """  # noqa: D400
+        if f is None:
+            return wraps_partial(Device.task, **kwargs)  # type: ignore[reportGeneralTypeIssues]
+
+        f.__belay__ = MethodMetadata(executer=TaskExecuter, kwargs=kwargs)
+        return f
+
+    @staticmethod
+    def thread(f=None, **kwargs) -> Callable:
+        """thread(f, *, minify=True, register=True, record=True)
+
+        Decorator that send code to device that spawns a thread when executed.
+
+        Can either be used as a staticmethod ``@Device.thread`` for marking methods in a subclass of ``Device``, or as a standard method ``@device.thread`` for marking functions to a specific ``Device`` instance.
+
+        Parameters
+        ----------
+        f: Callable
+            Function to decorate. Can only accept python literals as arguments.
+        minify: bool
+            Minify ``cmd`` code prior to sending.
+            Defaults to ``True``.
+        register: bool
+            Assign an attribute to ``self.thread`` with same name as ``f``.
+            Defaults to ``True``.
+        record: bool
+            Each invocation of the executer is recorded for playback upon reconnect.
+            Defaults to ``True``.
+        """  # noqa: D400
+        if f is None:
+            return wraps_partial(Device.task, **kwargs)  # type: ignore[reportGeneralTypeIssues]
+        f.__belay__ = MethodMetadata(executer=ThreadExecuter, kwargs=kwargs)
+        return f
+
     def _traceback_execute(
         self,
         src_file: Union[str, Path],
@@ -798,7 +738,7 @@ class Device:
         name: str,
         cmd: str,
         record: bool = True,
-    ) -> BelayReturn:
+    ):
         """Invoke ``cmd``, and reinterprets raised stacktrace in ``PyboardException``.
 
         Parameters
@@ -814,6 +754,10 @@ class Device:
         record: bool
             Record the call for state-reconstruction if device is accidentally reset.
             Defaults to ``True``.
+
+        Returns
+        -------
+            Correctly interpreted return value from executing code on-device.
         """
         src_file = str(src_file)
 
